from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any

from cachetools import LRUCache
from django.conf import settings
from django.template.backends.django import Template as DjangoTemplate
from django.template.loader import select_template

from django_bird.staticfiles import Asset

from .staticfiles import get_template_assets
from .templates import get_template_names


@dataclass(frozen=True, slots=True)
class Component:
    name: str
    template: DjangoTemplate
    assets: frozenset[Asset] = field(default_factory=frozenset)

    def render(self, context: dict[str, Any]):
        return self.template.render(context)

    @property
    def nodelist(self):
        return self.template.template.nodelist

    @classmethod
    def from_name(cls, name: str):
        template_names = get_template_names(name)
        template = select_template(template_names)
        assets = get_template_assets(template)
        return cls(name=name, template=template, assets=assets)


class Registry:
    def __init__(self, maxsize: int = 100):
        self._cache: LRUCache[str, Component] = LRUCache(maxsize=maxsize)

    def get_component(self, name: str) -> Component:
        try:
            return self._cache[name]
        except KeyError:
            component = Component.from_name(name)
            if not settings.DEBUG:
                self._cache[name] = component
            return component

    def clear(self) -> None:
        """Clear the component cache. Mainly useful for testing."""
        self._cache.clear()


components = Registry()



from __future__ import annotations

import logging
import re
from pathlib import Path

import pytest
from django.conf import settings
from django.template.backends.django import DjangoTemplates
from django.template.backends.django import Template as DjangoTemplate
from django.template.engine import Engine
from django.test import override_settings

from .settings import DEFAULT_SETTINGS

pytest_plugins = []


def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(**DEFAULT_SETTINGS, **TEST_SETTINGS)


TEST_SETTINGS = {
    "INSTALLED_APPS": [
        "django_bird",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [
                Path(__file__).parent / "templates",
            ],
            "OPTIONS": {
                "builtins": [
                    "django.template.defaulttags",
                    "django_bird.templatetags.django_bird",
                ],
                "loaders": [
                    "django_bird.loader.BirdLoader",
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ],
            },
        }
    ],
}


@pytest.fixture
def templates_dir(tmp_path):
    templates_dir = tmp_path / "templates"
    templates_dir.mkdir()
    return templates_dir


@pytest.fixture(autouse=True)
def override_templates_settings(templates_dir):
    with override_settings(
        TEMPLATES=[
            settings.TEMPLATES[0]
            | {
                "DIRS": [
                    *settings.TEMPLATES[0]["DIRS"],
                    templates_dir,
                ]
            }
        ]
    ):
        yield


@pytest.fixture
def create_bird_dir(templates_dir):
    def func(name):
        bird_template_dir = templates_dir / name
        bird_template_dir.mkdir(exist_ok=True)
        return bird_template_dir

    return func


@pytest.fixture
def create_bird_template(create_bird_dir):
    def func(name, content, sub_dir=None, bird_dir_name="bird"):
        bird_dir = create_bird_dir(bird_dir_name)
        if sub_dir is not None:
            dir = bird_dir / sub_dir
            dir.mkdir()
        else:
            dir = bird_dir
        template = dir / f"{name}.html"
        template.write_text(content)
        return template

    return func


@pytest.fixture
def create_bird_asset():
    def func(component_template: Path, content: str, asset_type: str):
        component_dir = component_template.parent
        component_name = component_template.stem
        asset_file = component_dir / f"{component_name}.{asset_type}"
        asset_file.write_text(content)
        return asset_file

    return func


@pytest.fixture
def create_template():
    def _create_template(template_file: Path) -> DjangoTemplate:
        engine = Engine(
            builtins=["django_bird.templatetags.django_bird"],
            dirs=[str(template_file.parent)],
            loaders=["django_bird.loader.BirdLoader"],
        )
        print(f"Engine dirs: {engine.dirs}")  # Debug print
        print(f"Looking for template: {template_file.name}")  # Debug print
        template = engine.get_template(template_file.name)
        backend = DjangoTemplates(
            {
                "NAME": "django",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "autoescape": True,
                    "debug": False,
                    "context_processors": [],
                },
            }
        )
        return DjangoTemplate(template, backend)

    return _create_template


@pytest.fixture
def normalize_whitespace():
    def func(text):
        # this makes writing tests much easier, as it gets rid of any
        # existing whitespace that may be present in the template file

        # multiple whitespace characters
        text = re.sub(r"\s+", " ", text)

        # after opening tag, including when there are attributes
        text = re.sub(r"<(\w+)(\s+[^>]*)?\s*>", r"<\1\2>", text)

        # before closing tag
        text = re.sub(r"\s+>", ">", text)

        # after opening tag and before closing tag
        text = re.sub(r">\s+<", "><", text)

        # immediately after opening tag (including attributes) or before closing tag
        text = re.sub(r"(<\w+(?:\s+[^>]*)?>)\s+|\s+(<\/\w+>)", r"\1\2", text)

        return text.strip()

    return func


@pytest.fixture(autouse=True)
def clear_registry():
    from django_bird.components import components

    components.clear()
    yield
    components.clear()



from __future__ import annotations

from pathlib import Path

import pytest

from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType
from django_bird.staticfiles import get_template_assets


class TestAsset:
    def test_hash(self):
        asset1 = Asset(Path("static.css"), AssetType.CSS)
        asset2 = Asset(Path("static.css"), AssetType.CSS)

        assert asset1 == asset2
        assert hash(asset1) == hash(asset2)

        assets = {asset1, asset2, Asset(Path("other.css"), AssetType.CSS)}

        assert len(assets) == 2

    def test_exists(self, tmp_path: Path):
        css_file = tmp_path / "test.css"
        css_file.touch()

        asset = Asset(css_file, AssetType.CSS)

        assert asset.exists() is True

    def test_exists_nonexistent(self):
        missing_asset = Asset(Path("missing.css"), AssetType.CSS)
        assert missing_asset.exists() is False

    @pytest.mark.parametrize(
        "path,expected",
        [
            (Path("static.css"), Asset(Path("static.css"), AssetType.CSS)),
            (Path("static.js"), Asset(Path("static.js"), AssetType.JS)),
            (
                Path("nested/path/style.css"),
                Asset(Path("nested/path/style.css"), AssetType.CSS),
            ),
            (
                Path("./relative/script.js"),
                Asset(Path("./relative/script.js"), AssetType.JS),
            ),
            (Path("UPPERCASE.CSS"), Asset(Path("UPPERCASE.CSS"), AssetType.CSS)),
            (Path("mixed.Js"), Asset(Path("mixed.Js"), AssetType.JS)),
        ],
    )
    def test_from_path(self, path, expected):
        assert Asset.from_path(path) == expected

    def test_from_path_invalid(self):
        with pytest.raises(ValueError):
            Asset.from_path(Path("static.html"))


class TestGetTemplateAssets:
    def test_with_assets(self, create_template, create_bird_template):
        template_file = create_bird_template(name="button", content="<button>")

        css_file = template_file.with_suffix(".css")
        js_file = template_file.with_suffix(".js")
        css_file.write_text("button { color: red; }")
        js_file.write_text("console.log('loaded');")

        template = create_template(template_file)
        assets = get_template_assets(template)

        assert len(assets) == 2
        assert Asset(css_file, AssetType.CSS) in assets
        assert Asset(js_file, AssetType.JS) in assets

    @pytest.mark.parametrize(
        "suffix,content,expected_type",
        [
            (".css", "button { color: red; }", AssetType.CSS),
            (".js", "console.log('loaded');", AssetType.JS),
        ],
    )
    def test_partial_assets(
        self, suffix, content, expected_type, create_template, create_bird_template
    ):
        template_file = create_bird_template(name="button", content="<button>")

        file = template_file.with_suffix(suffix)
        file.write_text(content)

        template = create_template(template_file)

        assets = get_template_assets(template)

        assert len(assets) == 1
        assert Asset(file, expected_type) in assets

    def test_no_assets(self, create_template, create_bird_template):
        template_file = create_bird_template(name="button", content="<button>")

        template = create_template(template_file)

        assets = get_template_assets(template)

        assert len(assets) == 0

    def test_custom_component_dir(self, create_template, create_bird_template):
        template_file = create_bird_template(
            name="button", content="<button>", sub_dir="components"
        )

        css_file = template_file.with_suffix(".css")
        css_file.write_text("button { color: red; }")

        template = create_template(template_file)

        assets = get_template_assets(template)

        assert len(assets) == 1
        assert Asset(css_file, AssetType.CSS) in assets



# pyright: reportAny=false
from __future__ import annotations

from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override
from django_bird.loader import BirdLoader
from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType

CSS_TAG = "bird:css"
JS_TAG = "bird:js"


def do_asset(parser: Parser, token: Token) -> AssetNode:
    bits = token.split_contents()
    asset_type = parse_asset_type(bits)
    return AssetNode(asset_type)


def parse_asset_type(bits: TagBits) -> AssetType:
    if len(bits) < 1:
        msg = "bird:assets tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    tag_name = bits[0]

    try:
        asset_type = tag_name.split(":")[1]
        match asset_type:
            case "css":
                return AssetType.CSS
            case "js":
                return AssetType.JS
            case _:
                raise ValueError(f"Unknown asset type: {asset_type}")
    except IndexError as e:
        raise ValueError(f"Invalid tag name: {tag_name}") from e


class AssetNode(template.Node):
    def __init__(self, asset_type: AssetType):
        self.asset_type = asset_type

    @override
    def render(self, context: Context) -> str:
        template = context.template
        if template is None:
            return ""
        loaders = template.engine.template_loaders
        for loader in loaders:
            if isinstance(loader, BirdLoader):
                assets = loader.asset_registry.get_assets(self.asset_type)
                return self._render_assets(assets)

        raise RuntimeError("BirdLoader not found in template loaders")

    def _render_assets(self, assets: set[Asset]) -> str:
        if not assets:
            return ""

        if self.asset_type == AssetType.CSS:
            tags = (
                f'<link rel="stylesheet" href="{asset.path}">'
                for asset in sorted(assets, key=lambda a: a.path)
            )
        else:  # JS
            tags = (
                f'<script src="{asset.path}"></script>'
                for asset in sorted(assets, key=lambda a: a.path)
            )

        return "\n".join(tags)



from __future__ import annotations

import pytest
from django.template.base import Node
from django.template.base import NodeList
from django.template.base import Template
from django.template.context import Context
from django.template.engine import Engine
from django.template.loader import get_template

from django_bird.loader import BIRD_TAG_PATTERN
from django_bird.loader import BirdLoader
from django_bird.params import Params
from django_bird.templatetags.tags.bird import BirdNode


@pytest.mark.parametrize(
    "template_content,expected_matches",
    [
        ("{% bird button %}", ["button"]),
        ("{% bird alert %}{% bird button %}", ["alert", "button"]),
        ("{% bird 'button' %}", ["button"]),
        ('{% bird "button" %}', ["button"]),
        ("{%bird button%}", ["button"]),
        ("{%    bird    button    %}", ["button"]),
        ("{% endbird button %}", []),
        ("{% birds button %}", []),
        ("<bird:button>", []),
        ("{% extends 'base.html' %}{% bird button %}", ["button"]),
    ],
)
def test_bird_tag_pattern(template_content, expected_matches):
    matches = [
        m.group(1).strip("'\"") for m in BIRD_TAG_PATTERN.finditer(template_content)
    ]
    assert matches == expected_matches


@pytest.mark.parametrize(
    "template_name",
    (
        "with_bird.html",
        "without_bird.html",
    ),
)
def test_render_template(template_name):
    context = {
        "title": "Test Title",
        "content": "Test Content",
    }

    template = get_template(template_name)
    rendered = template.render(context)

    assert rendered


def test_asset_registry(
    create_bird_template, create_bird_asset, create_template, templates_dir
):
    alert = create_bird_template("alert", '<div class="alert">{{ slot }}</div>')
    create_bird_asset(alert, ".alert { color: red; }", "css")
    create_bird_asset(alert, "console.log('alert');", "js")

    badge = create_bird_template("badge", "<span>{{ slot }}</span>")
    create_bird_asset(badge, ".badge { color: blue; }", "css")
    create_bird_asset(badge, "console.log('badge');", "js")

    button = create_bird_template("button", "<button>{{ slot }}</button>")
    create_bird_asset(button, ".button { color: blue; }", "css")
    create_bird_asset(button, "console.log('button');", "js")

    base_path = templates_dir / "base.html"
    base_path.write_text("""
        <html>
        <head>
            <title>Test</title>
            {% bird:css %}
        </head>
        <body>
            {% bird alert %}Base Alert{% endbird %}
            {% block content %}{% endblock %}
            {% bird:js %}
        </body>
        </html>
    """)

    include_path = templates_dir / "include.html"
    include_path.write_text("""
        {% bird badge %}Active{% endbird %}
    """)

    child_path = templates_dir / "child.html"
    child_path.write_text("""
        {% extends 'base.html' %}
        {% block content %}
            {% bird button %}Click me{% endbird %}
            {% include 'include.html' %}
        {% endblock %}
    """)

    template = create_template(child_path)

    engine = template.template.engine
    loader = engine.template_loaders[0]

    components = loader.asset_registry.components

    assert len(components) == 3


@pytest.mark.parametrize(
    "node,expected_count",
    [
        (Template("{% bird button %}Click me{% endbird %}"), 1),
        (BirdNode(name="button", params=Params([]), nodelist=None), 1),
        (
            BirdNode(
                name="button",
                params=Params([]),
                nodelist=NodeList(
                    [BirdNode(name="button", params=Params([]), nodelist=None)],
                ),
            ),
            1,
        ),
        (type("NodeWithNoneNodelist", (Node,), {"nodelist": None})(), 0),
        (Template("{% bird button %}{% bird button %}{% endbird %}{% endbird %}"), 1),
    ],
)
def test_scan_for_components(
    node, expected_count, create_bird_template, create_bird_asset
):
    button = create_bird_template("button", "<button>Click me</button>")
    create_bird_asset(button, ".button { color: blue; }", "css")
    create_bird_asset(button, "console.log('button');", "js")

    loader = BirdLoader(Engine.get_default())
    context = Context()

    loader._scan_for_components(node, context)

    assert len(loader.asset_registry.components) == expected_count



from __future__ import annotations

import pytest
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.context import Context
from django.template.engine import Engine
from django.template.exceptions import TemplateSyntaxError

from django_bird.staticfiles import AssetType
from django_bird.templatetags.tags.asset import CSS_TAG
from django_bird.templatetags.tags.asset import JS_TAG
from django_bird.templatetags.tags.asset import AssetNode
from django_bird.templatetags.tags.asset import do_asset


class TestTemplateTag:
    @pytest.mark.parametrize(
        "tag,expected",
        [
            (CSS_TAG, AssetType.CSS),
            (JS_TAG, AssetType.JS),
        ],
    )
    def test_asset_type(self, tag, expected):
        token = Token(TokenType.BLOCK, tag)
        parser = Parser([])
        node = do_asset(parser, token)
        assert node.asset_type == expected

    def test_missing_tag_name(self):
        token = Token(TokenType.BLOCK, "")
        parser = Parser([])
        with pytest.raises(TemplateSyntaxError):
            do_asset(parser, token)

    @pytest.mark.parametrize(
        "tag",
        ["bird:jsx", "birdcss"],
    )
    def test_invalid_tag_name(self, tag):
        token = Token(TokenType.BLOCK, tag)
        parser = Parser([])
        with pytest.raises(ValueError):
            do_asset(parser, token)

    def test_template_inheritence(
        self, create_bird_template, create_bird_asset, create_template, templates_dir
    ):
        alert = create_bird_template("alert", '<div class="alert">{{ slot }}</div>')
        alert_css = create_bird_asset(alert, ".alert { color: red; }", "css")
        alert_js = create_bird_asset(alert, "console.log('alert');", "js")

        button = create_bird_template("button", "<button>{{ slot }}</button>")
        button_css = create_bird_asset(button, ".button { color: blue; }", "css")
        button_js = create_bird_asset(button, "console.log('button');", "js")

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        child_path = templates_dir / "child.html"
        child_path.write_text("""
            {% extends 'base.html' %}
            {% block content %}
                {% bird button %}Click me{% endbird %}
            {% endblock %}
        """)

        template = create_template(child_path)

        rendered = template.render({})

        assert f'<link rel="stylesheet" href="{alert_css}">' in rendered
        assert f'<link rel="stylesheet" href="{button_css}">' in rendered
        assert f'<script src="{alert_js}"></script>' in rendered
        assert f'<script src="{button_js}"></script>' in rendered

    def test_with_no_assets(
        self, create_bird_template, create_bird_asset, create_template, templates_dir
    ):
        create_bird_template("alert", '<div class="alert">{{ slot }}</div>')

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        template = create_template(base_path)

        rendered = template.render({})

        assert '<link rel="stylesheet" href="' not in rendered
        assert '<script src="' not in rendered

    def test_component_render_order(
        self, create_bird_template, create_bird_asset, create_template, templates_dir
    ):
        first = create_bird_template("first", "<div>First: {{ slot }}</div>")
        first_css = create_bird_asset(first, ".first { color: red; }", "css")
        first_js = create_bird_asset(first, "console.log('first');", "js")

        second = create_bird_template("second", "<div>Second: {{ slot }}</div>")
        second_css = create_bird_asset(second, ".second { color: red; }", "css")
        second_js = create_bird_asset(second, "console.log('second');", "js")

        template_path = templates_dir / "test.html"
        template_path.write_text("""
            <html>
            <head>
                {% bird:css %}
            </head>
            <body>
                <div>
                    {% bird first %}One{% endbird %}
                    <p>Some content</p>
                    {% bird second %}Two{% endbird %}
                </div>
                {% bird:js %}
            </body>
            </html>
        """)

        template = create_template(template_path)

        rendered = template.render({})

        head_end = rendered.find("</head>")
        assert f'<link rel="stylesheet" href="{first_css}">' in rendered[:head_end]
        assert f'<link rel="stylesheet" href="{second_css}">' in rendered[:head_end]

        body_start = rendered.find("<body")
        assert f'<script src="{first_js}"></script>' in rendered[body_start:]
        assert f'<script src="{second_js}"></script>' in rendered[body_start:]

    def test_birdloader_required(self, templates_dir, create_template):
        template_path = templates_dir / "test.html"
        template_path.write_text("""
            <html>
            <head>
                {% bird:css %}
            </head>
            <body>
                {% bird:js %}
            </body>
            </html>
        """)

        engine = Engine(
            builtins=["django_bird.templatetags.django_bird"],
            dirs=[str(templates_dir)],
            loaders=["django.template.loaders.filesystem.Loader"],
        )
        template = engine.get_template("test.html")

        with pytest.raises(RuntimeError):
            template.render(Context({}))


class TestNode:
    def test_no_template(self):
        node = AssetNode(AssetType.CSS)
        context = Context({})
        assert node.render(context) == ""



from __future__ import annotations

import pytest
from django.template.backends.django import Template
from django.template.exceptions import TemplateDoesNotExist
from django.test import override_settings

from django_bird.components import Component
from django_bird.components import Registry
from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType


class TestComponent:
    def test_from_name_basic(self, create_bird_template):
        create_bird_template("button", "<button>Click me</button>")

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert comp.assets == set()
        assert isinstance(comp.template, Template)
        assert comp.render({}) == "<button>Click me</button>"

    def test_from_name_with_assets(self, create_template, create_bird_template):
        template_file = create_bird_template("button", "<button>Click me</button>")
        create_template(template_file)

        css_file = template_file.with_suffix(".css")
        js_file = template_file.with_suffix(".js")
        css_file.write_text("button { color: red; }")
        js_file.write_text("console.log('loaded');")

        comp = Component.from_name("button")

        assert len(comp.assets) == 2
        assert Asset(css_file, AssetType.CSS) in comp.assets
        assert Asset(js_file, AssetType.JS) in comp.assets

    @pytest.mark.parametrize(
        "asset_suffix,asset_content,expected_asset_type",
        [
            (".css", "button { color: red; }", AssetType.CSS),
            (".js", "console.log('loaded');", AssetType.JS),
        ],
    )
    def test_from_name_with_partial_assets(
        self,
        asset_suffix,
        asset_content,
        expected_asset_type,
        create_template,
        create_bird_template,
    ):
        template_file = create_bird_template("button", "<button>Click me</button>")
        create_template(template_file)

        file = template_file.with_suffix(asset_suffix)
        file.write_text(asset_content)

        comp = Component.from_name("button")

        assert len(comp.assets) == 1
        assert Asset(file, expected_asset_type) in comp.assets


class TestRegistry:
    @pytest.fixture
    def registry(self):
        return Registry(maxsize=2)

    def test_get_component_caches(self, registry, create_bird_template):
        create_bird_template(name="button", content="<button>Click me</button>")

        component1 = registry.get_component("button")
        component2 = registry.get_component("button")

        assert component1 is component2

    def test_lru_cache_behavior(self, registry, create_bird_template):
        create_bird_template(name="button1", content="1")
        create_bird_template(name="button2", content="2")
        create_bird_template(name="button3", content="3")

        button1 = registry.get_component("button1")
        button2 = registry.get_component("button2")
        button3 = registry.get_component("button3")

        new_button1 = registry.get_component("button1")
        assert new_button1 is not button1

        cached_button2 = registry.get_component("button2")
        assert cached_button2.name == button2.name
        assert cached_button2.render({}) == button2.render({})

        cached_button3 = registry.get_component("button3")
        assert cached_button3.name == button3.name
        assert cached_button3.render({}) == button3.render({})

    def test_component_not_found(self, registry):
        with pytest.raises(TemplateDoesNotExist):
            registry.get_component("nonexistent")

    def test_cache_with_debug(self, registry, create_bird_template):
        create_bird_template(name="button", content="<button>Click me</button>")

        assert len(registry._cache) == 0

        with override_settings(DEBUG=True):
            registry.get_component("button")

        assert len(registry._cache) == 0

        registry.get_component("button")

        assert len(registry._cache) == 1



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

from django.template.backends.django import Template as DjangoTemplate

from ._typing import override

if TYPE_CHECKING:
    from django_bird.components import Component


class AssetType(Enum):
    CSS = "css"
    JS = "js"

    @property
    def ext(self):
        return f".{self.value}"


@dataclass(frozen=True, slots=True)
class Asset:
    path: Path
    type: AssetType

    @override
    def __hash__(self) -> int:
        return hash((str(self.path), self.type))

    def exists(self) -> bool:
        return self.path.exists()

    @classmethod
    def from_path(cls, path: Path) -> Asset:
        match path.suffix.lower():
            case ".css":
                asset_type = AssetType.CSS
            case ".js":
                asset_type = AssetType.JS
            case _:
                raise ValueError(f"Unknown asset type for path: {path}")
        return cls(path=path, type=asset_type)


def get_template_assets(template: DjangoTemplate):
    assets: set[Asset] = set()
    template_path = Path(template.template.origin.name)
    for asset_type in AssetType:
        asset_path = template_path.with_suffix(asset_type.ext)
        if asset_path.exists():
            asset = Asset.from_path(asset_path)
            assets.add(asset)
    return frozenset(assets)


@dataclass
class ComponentAssetRegistry:
    components: set[Component] = field(default_factory=set)

    def register(self, component: Component) -> None:
        self.components.add(component)

    def get_assets(self, asset_type: AssetType) -> set[Asset]:
        assets: set[Asset] = set()
        for component in self.components:
            assets.update(a for a in component.assets if a.type == asset_type)
        return assets
